#!/usr/bin/env python

#
# provides the object new, which is a child of class output_plugin
#

import sys
import os
import re
import popen2
import signal
import fcntl, FCNTL

import mp3meta
import input_plugin
import ui
import multiplex
import player
import config

from debug import debug

PLUGIN_TYPE='in'

default_config = {
    "mpg123_executable": "mpg123"
}

config.set_defaults("mp3ctl", default_config)

class new(input_plugin.new):
    """provides an input_plugin capable of playing mp3 files"""

    def __init__(self):

        self.file_match = re.compile(r"\.mp[23]$")
        self.status_match = re.compile(\
            r"Frame#\s*(\d+) \[\s*(\d+)\], Time: (\d\d):(\d\d).(\d\d) " \
            "\[(\d\d):(\d\d).(\d\d)\],")
        self.finished_match = re.compile(r"Decoding of")
        # we only want to handle ~ 1/50 frames to keep processing down

        self.skip_destination = 0
        self.mpg123_executable = config.get("mp3ctl", "mpg123_executable")

        debug(2, "mp3ctl using executable: " + self.mpg123_executable)
        
        self._reset_state()

        # call out parent's constuctor
        input_plugin.new.__init__(self)

        debug(3, "mp3ctl init")

    def filename_to_meta(self, path):
        """ given a file pathname, return the artist, title etc """

        ret = {}
        try:
            f = mp3meta.parse_filename(os.path.basename(path))            
            if  not f:
                f = mp3meta.parse_id3(path)
                ret['artist'] = mp3meta.nice_case(f['artist'])
                ret['title'] = mp3meta.nice_case(f['title'])
            else:
                ret['artist'] = mp3meta.nice_case(f['artist'])
                ret['title'] = mp3meta.nice_case(f['title'])

        except: pass
            
        if not ret:
            ret['artist'] = os.path.basename(path).replace(".mp3", "")
            ret['title'] = ""

        return ret

    def hook(self, fd):
        """this function is called when there's action on our current FD."""

        # read a single line (up to 80 bytes) from our pipe, and assert
        # that it has data - if it doesn't, why has hook been called?

        # FIXME: actually, if we were to listen to POLLHUP too, we might
        # be called with no data to read; need to rework this

        # read up to 8 frames before returning (don't want to hog CPU)
        
        line = os.read(self.fd, 80)

        # now process the line
        if self._parse_status(line):

            if self.rem_frame == 0:
                # the song has now finished.
                debug(1, "foo!")
                self._finish()
                return
            
            else:
                # normally update the status display, and re-schedule
                # us to receive events on the FD

                self.secs = self.cur_mins * 60 + self.cur_secs
                ui.new_time()                
                ui.refresh()
        else:
            # not a usual status line, check what's going on
            self._parse_other(line)


    def _finish(self):
        """ song has finished, clean up, update the user interface, etc """

        # now clear our internal data, child, etc

        self.stop()
        
        self._reset_state()
        
        # signal to the player that we need advance
        player.finished()
        
    def _parse_other(self, line):
        """ check for other things, like a song finishing abruptly etc """

        # unfortunately most of these are errors resulting from mpg123 being
        # unable to handle some mp3s. so when errors are reported, we kick into
        #

        if re.match("^Free format not supported", line):
            # mpg123 really doesn't like these. switch players and kick
            # back in at start
            self._switch_and_kick()

        # theoretically, this should be caught by other errors
        elif self.finished_match.search(line):
            # song has finished, although not necessarily at the end.
            self._switch_and_kick(resume=1)

        elif re.match("^Illegal Audio-MPEG-Header", line):
            self._switch_and_kick(resume=1)

        elif re.match("^big_values too large!", line):
            self._switch_and_kick(resume=1)
        
            
        # this is actually a joke, as you can't match on that. but
        # it's happened before # unfortunately :-)

        # elif re.match("^Segmentation fault", line):
        # self._switch_and_kick(resume=1)

    def _switch_and_kick(self, resume=0):
        """ switch our backend to mpg321 to recover from
        errors. resume indicates if we try playing from the start
        again. this only is active for the current file, and if rem_frame
        is < 100, we just move to the next. """

        debug(2, "switch&kick")

        if self.mpg123_executable=="mpg321":
            self.stop()
            player.finished()

        if (self.rem_frame < 100 and self.frame == 0):
            # we've already kicked or there's not enough left to bother.
            # just start the next song - but if frame=0 then we haven't decoded
            # any frames, and rem_frame will be wrong
            
            self.stop()
            player.finished()
            return

        debug("kicking to mpg321 due to error")

        new_player = "mpg321"
        
        file = self.file            
        self._stop()
        self.switched += 1
        
        if resume:
            self._play(file, self.frame, new_player)
        else:
            self._play(file, player=new_player)
            
    def _parse_status(self, line):
        """takes a line of output from mpg123 and fill in the status info
        returns true on success."""

        m = self.status_match.search(line)

        if m:
            # a successful status line from mpg123
            self.frame      = int(m.group(1))
            self.rem_frame  = int(m.group(2))
            self.cur_mins  = int(m.group(3))
            self.cur_secs   = int(m.group(4))
            self.cur_msecs   = int(m.group(5))
            self.rem_mins  = int(m.group(6))
            self.rem_secs   = int(m.group(7))
            self.rem_msecs   = int(m.group(8))

            # processed successfully
            return 1

        else:
            # not processed
            return 0
        
    def validate_filename(self, file):
        """given a filename, returns if it is able to parse this file"""

        if self.file_match.search(file):
            # the file is one of ours
            return 1
        else:
            # the file isn't ours
            return 0

    def validate_content(self, file):
        """given a filename, checks content - is it able to parse file?"""

        # this gets a bit trickier. using the mp3.py module, we
        # attempt to parse the file. if we catch an exception, chances
        # are this isn't an mp3 file.

        debug("broken atm"); assert(0)
        
        f = mp3meta.file(file)

        if f.valid():
            # this file *appears* to be an mp3 file
            return 1
        else:
            # whoops, must be another file
            return 0

    def locked(self):
        """ we're locked if there's no current file to play """

        return self.file == None

    def is_playing(self):
        """ is there a file currently playing? """

        return self.fd != None

    def play(self, file=None, frame=0):
        """plays a filename. """
        self._play(file, frame)

    def _play(self, file=None, frame=0, player=None):

        if player==None: player=self.mpg123_executable
        
        self.paused = 0

        self.file = file
        
        (pin, pout) = os.pipe()

        pid = os.fork()

        if pid == 0:
            # we're the child


            # set the session ID, so both mpg123 and this process are killed
            # when a stop is issued
            
            os.setsid()

            # figure out arguments
            args = [ player, "-v", "-k", str(frame), "-b 512", \
                     self.file ]
                     
            # now dup the fd and load mpg123 on top

            os.dup2(pout, os.sys.stderr.fileno())
            os.execvp(player, args)
            
        else:
            # we're the parent
            self.fd = pin
            self.child_pid = pid
            
            # renice the player
            os.system("renice -5 -g %d > /dev/null 2>&1" % self.child_pid)
            
            multiplex.add_watch(self.fd, multiplex.POLLIN, self.hook)

            # now we have to try parse the first frame. we block on this to
            # prevent returning to the main app before mpg123 has started up.

            #  FIXME: possible deadlock if mpgx doesn't start.
            self.hook(self.fd)

            # now that we've blocked once, change to non-blocking mode
            fcntl.fcntl(self.fd, FCNTL.F_SETFL, FCNTL.O_NONBLOCK)

        self.status = "playing"
        ui.new_status()
        ui.refresh()
    
    def _stop(self):
        """stops any current playing, may be called by pause too"""

        if self.locked(): return

        # take out the mpg123 process
        # update: sigkill to prevent an errant mpg123 process from preventing
        # further songs
        os.kill(-self.child_pid, signal.SIGKILL)
        # wait for all children to be removed from P.T.
        os.waitpid(-self.child_pid, 0) 

        # delete our entry in the poll_table before fd is reset
        multiplex.del_watch(self.fd)

        # close our pipe
        os.close(self.fd)

        self.switched = 0
        self.file=None

    def stop(self):
        """ calls _stop and then resets everything, including user interface """
        
        self._stop()

        self._reset_state()

        # no current filename, fd etc
        self.status = "stopped"
        ui.new_status()
        ui.refresh()

    def _reset_state(self):
        
        self.paused = 0
        if hasattr(self, 'file'):
            self.oldfile = self.file
        self.file = None
        self.fd = None
        self.switched = 0
        self.frame = 0
        self.rem_frame = 0
        self.skip_destination=0

        self.status = "stopped"

    def pause(self):
        """stop the current song playing, but remember where it is"""

        if self.paused == 1:
            # if we're currently paused, unpause
            self.play(self.paused_file, self.frame)
            self.paused = 0
            self.paused_file = None
            
        elif not self.locked():
            self.status = "paused"
            ui.new_status()
            ui.refresh()
            # we're not paused - pause.

            self.paused_file = self.file

            self._stop()
            self.paused = 1

    def fwd(self):
        """jump forward in a track"""

        # rewrote this to have a set delay before skipping - works better
        # on slower machines.

        # we check here, but we must also check in the hook

        if self.locked(): return

        if self.rem_frame < (self.skip_destination + 300):
            return

        self.skip_destination += 300

        secs = self.skip_destination/38 # rough, very rough! :)

        ui.notice("skipping forward ~ %.2d:%.2d" % (secs/60, secs%60))
        ui.refresh()

        multiplex.set_timer("skip", 1, lambda s=self: s._fwd())

    def _fwd(self):

        if self.locked():
            self.skip_destination = 0
            return
        if self.rem_frame < self.skip_destination:
            # too far? new song
            self._finish()
            return

        file  = self.file
        frame = self.frame

        if self.is_playing(): self._stop()
        self.play(file, frame + self.skip_destination)

        self.skip_destination = 0

    def back(self):
        """jump back in a track"""

        if self.locked():
            return

        foo = self.frame - (self.skip_destination + 300)
        if foo < 0:
            self.skip_destination += 300 + foo
        else:
            self.skip_destination += 300

        secs = self.skip_destination/38 # rough, very rough! :)
                
        ui.notice("skipping back ~ %.2d:%.2d" % (secs/60, secs%60))
        ui.refresh()

        multiplex.set_timer("skip", 1, lambda s=self: s._back())

    def _back(self):

        if self.locked():
            self.skip_destination = 0
            return            
        if self.frame - self.skip_destination < 0:
            self.skip_destination = self.frame
        
        file  = self.file
        frame = self.frame

        if self.is_playing(): self._stop()
        self.play(file, frame - self.skip_destination)

        self.skip_destination = 0

    def message(self, msg):

        if msg == "switch_player":
            if self.mpg123_executable=="mpg123":
                ui.notice("switching to mpg321 (madplay)")
                ui.refresh()
                self.mpg123_executable="mpg321"
            else:
                ui.notice("switching to mpg123 (standard)")
                ui.refresh()
                self.mpg123_executable="mpg123"


    def get_pos(self):
        """ return state in song, if possible. doesn't need to be
        implemented """
        return self.frame

    def set_pos(self, pos):
        file = self.file

        if self.is_playing(): self._stop()
        self.play(file, pos)
            
            



 
