#!/usr/bin/env python
#
# multiplex.py - provides an interface to the main loop, where modules can
#                provide fds to watch, etc. this module also provides a timer
#                interface.
#
# started 8/8/01 by Damien Elmes <resolve@repose.cx>
# $Id: multiplex.py,v 1.11 2001/09/20 17:59:53 resolve Exp resolve $
#


import sys

from debug import *
# we import select in order to provide stuff like POLLIN
from select import *

import time

poll_table = {}
timers = []
idles = []

# sets the resolution of timers. 1sec is a safe choice, as time.time() is only
# accurate to that on some systems. clock() could be used to implement a timer
# with a higher resolution, although it's always going to stuff up if one of
# the hooks takes a long time.

# using 0.1sec
idle_granularity = 100

def add_watch(fd, mask, hook):
    """ add an entry to the watch table. mask takes the same values as
    select.poll.register(..., mask) """

    poll_table[fd] = (mask, hook)

def del_watch(fd):
    """ delete an entry from the poll table """

    try:
        del poll_table[fd]
    except KeyError: pass


def add_timer(label, secs, hook):

    target = time.time() + secs

    timers.append({ 'label': label, 'target': target, 'hook': hook })
    

def set_timer(label, secs, hook):
    """ call hook when secs have elapsed - label is used for managing the
    timer. if label is already used, then update the details.

    note: you can't use set_timer to renew a called timer. create a new one"""

    target = time.time() + secs

    for unit in timers:
        if label == unit['label']:
            unit['target'] = target
            unit['hook'] = hook
            return

    # not existing, add new one
    timers.append({ 'label': label, 'target': target, 'hook': hook })

def del_timer(label):
    """ delete a timer - this should no-op if the timer has been called"""

    def filt(unit, label=label):
        if label == unit['label']:
            return 0
        return 1

    global timers
    timers = filter(filt, timers)

def add_idle(func):
    """ add an idle task to be called. if the task returns 0, it will be
    removed """

    idles.append(func)

def main_loop():
    """ poll and call any hooks where appropriate. generally you'd call this
    from your main function. this function also checks to make sure to make
    sure there are no outstanding timers waiting. """

    while 1:

        # setup a poll object with our current list of FDs
        polls = poll()

        for x in poll_table.items():
            (fd, (mask, _)) = x
            polls.register(fd, mask)

        try:
            pending_fds = polls.poll(idle_granularity)
        except error:
            debug(1, "select.error on poll, this is probably a resized xterm.")
            debug(1, "don't do that :-) exiting..")

            sys.exit(1)


        # iterate through all the fds with action on em.
        for x in pending_fds:
            (fd, mask) = x
            poll_table[fd][1](fd)

        # check to see if any timers need to be called
        cur_time = time.time()

        def filt(t, cur_time=cur_time):
            if t['target'] < cur_time:
                t['hook']()
                return 0
            return 1

        global timers
        timers = filter(filt, timers)

        # run any idle tasks, ones that return 0 are removed
        def filt(func):
            return func()

        global idles
        idles = filter(filt, idles)

        


                
        

    

    






    
