#!/usr/bin/env python
#
# player.py - controls playing of music files. supports modules for playing
#             different types of files.
# $Id: player.py,v 1.1 2001/09/20 19:35:56 resolve Exp resolve $
# started 5/8/01 by Damien Elmes <rapport@repose.cx>
#

import atexit
import os, sys
import time

import playlist
import ui
import multiplex

import config

from debug import debug

plugins = []
current_plugin = None

config_defaults = {
    "persist_file" : os.path.expanduser("~/.muse/persist.dat"),
    "persist" : 1
    }

config.set_defaults("player", config_defaults)

PERSIST_FILE=config.get("player", "persist_file")

def __init__():
    """ called once when the module is loaded """
    atexit.register(stop)

def playing():
    if current_plugin == None:
        return 0
    if not current_plugin.is_playing():
        return 0
    
    return 1
    
def add_plugin(plugin):
    plugins.append(plugin.new())

def have_handler(file):
    """ iterate through our modules, and check to make sure a file is playable
    by at least one of them """

    found = 0
    for x in plugins:
        if x.validate_filename(file):
            found = 1

    return found

def filename_to_meta(file):
    """ iterate through modules, returning the meta information from the first
    available module """

    for x in plugins:
        if x.validate_filename(file):
            return x.filename_to_meta(file)

    return None

def play(pl_cur):

    # determine which player is capable of playing this file
    global playlist_cur
    playlist_cur = pl_cur

    if pl_cur == None:
        # empty playlist object, which means there is nothing to play
        # (or something has gone very wrong :-)
        debug(2, "Warning, play() got empty playlist object")
        ui.notice("No songs to play!")
        return

    file = pl_cur['path']

    if not os.path.isfile(file):
        debug(2, "File passed to play() does not exist")
        return

    debug(2, "playing " + pl_cur['file'])


    # we're playing the file, so the playlist is clean again

    playlist.list.__dict__['dirty'] = 0

    if playing(): stop()

    # debug("playing " + os.path.basename(file))

    for x in plugins:
        # fixme: add config option to validate via content
        if x.validate_filename(file):
            # set the current module for playing
            global current_plugin
            current_plugin = x

            if config.get("player", "persist"):

                # update the persist file so that if we're shut down, we can
                # continue from this song. we store the size of the playlist for
                # some very basic consistancy checking.

                # doing this now is probably better than after we start playing,
                # to prevent the poor laptop hard drive from skipping around.

                path = os.path.dirname(sys.argv[0])
                f = open(os.path.join(path,PERSIST_FILE), 'w')

                f.write("%d\n" % (len(playlist.list),)) # playlist length
                f.write("%s\n" % (file,)) # playing file

                # i run ext3 on my car player, and don't bother shutting down,
                # so it's nice to flush any buffers after we do this - but we
                # wait about 8 seconds so that we don't slow down song skipping
                # terribly.

                f.flush()
                os.fsync (f.fileno())

                f.close()
                

            #multiplex.set_timer("sync", 8, lambda os=os: os.system("sync&"))

            # start playing.. :)
            
            ui.new_song() 

            current_plugin.play(file)

def start(persist=1):
    """ called by the main function, just starts playing the first song in
    the playlist, if there is one
    note: if the file /muse/persist.dat exists and is valid, then it will
    try to resume where we left off. if passed persist=0, then we move
    back to the start of the playlist """

    if playing(): current_plugin.stop()
    
    if len(playlist.list) == 0:
        debug(1, "can't start on empty playlist")
        return

    playlist.list.set_pos(0)
    if persist:
        # resume last pos
        try:
            f = open(PERSIST_FILE)

            plen = int(f.readline().strip())
            file = f.readline().strip()

            try:
                pos = f.readline().strip()
            except:
                pos = None

        except:
            f = None

        if f and plen == len(playlist.list):
            # good, the playlists appear to be the same

            if playlist.list.jump_to_path(file):
                # found the file
                debug(2, "resuming at file %s" % (os.path.basename(file),))

    play(playlist.list.current())

def stop():
    if current_plugin: current_plugin.stop()

def pause():
    if current_plugin: current_plugin.pause()

def fwd():
    if current_plugin: current_plugin.fwd()

def back():
    if current_plugin: current_plugin.back()


def clear_and_jump(func=None):
    
    """ clear the playlist, run function, and start playing the current
    playlist item """
    
    playlist.queue.clear()
    if func: func()
    playlist.queue.append(playlist.list.current())

    # Fixme: clear cur should always exist, no?
    play_currentq()

def next():
    ui.notice("skipped to the next track")
    clear_and_jump(playlist.list.next)

def next_with_queue():
    if len(playlist.queue)<=1:
        # if there is nothing left in the queue
        playlist.queue.clear()
        if not playlist.list.dirty: # advance if the user hasn't touched stuff
            playlist.list.next()
        ui.notice("playing current")
        playlist.queue.append(playlist.list.current())        
    else:
        playlist.queue.push()
        ui.notice("playing next song in queue (%d left)" % \
                  (len(playlist.queue)-1))

    play_currentq()

def jump_to_field(field=None, search=None):

    if playlist.list.jump_to_field(field, search):
        if field==search==None:
            ui.notice("continued search")
        else:
            ui.notice("jumped to %s: %s" % (field, search))
        ui.refresh()
    else:
        ui.notice("no match for %s: %s" % (field, search))
        ui.refresh()
        return
    
    clear_and_jump()

def add_to_queue():
    playlist.queue.append( playlist.list.current() )

def play_currentq():
    play( playlist.queue.current() )

def play_current():
    playlist.queue.clear()
    playlist.queue.append(playlist.list.current())

    play( playlist.queue.current() )
                          
def prev():
    ui.notice("skipped to the previous track")
    clear_and_jump(playlist.list.prev)

def next_artist():
    ui.notice("skipped to next artist")
    clear_and_jump(playlist.list.next_artist)
    
def prev_artist():
    ui.notice("skipped to previous artist")
    clear_and_jump(playlist.list.prev_artist)    

def next_dir():
    clear_and_jump(playlist.list.next_dir)
    ui.notice("ND: " + \
              os.path.split(playlist.queue.current()['dir'])[1])
    
def prev_dir():
    clear_and_jump(playlist.list.prev_dir)
    ui.notice("PD: " + \
              os.path.split(playlist.queue.current()['dir'])[1])

def current():
    return queue.current()
    
def finished():
    """ called from one of our plugins, indicating that a song has finished """

    
    next_with_queue()

def quit():
    debug(1, "player.quit() called")
    sys.exit(0)

    

# hack, FIXME do it better
def message(msg):
    if current_plugin: current_plugin.message(msg)

# load init stuff first time around
__init__()


def status():
    if current_plugin:
        return current_plugin.status

def secs():
    if current_plugin:
        return current_plugin.secs    





 

