#!/usr/bin/env python
#
# playlist.py - provides a playlist class, a simple media object, and two
#               playlist instances
#
# $Id: playlist.py,v 1.1 2001/09/20 19:37:13 resolve Exp resolve $
# started 17/8/01 by Damien Elmes <rapport@repose.cx>
#

# we provide two playlist instances for use in muse
# these are the list' - a list of all files available for playing
# and the 'queue', a list of files which have been queued up. when the queue
# is exhausted, playing will resume at the current playlist position.

# NB: these are defined after the playlist class

from debug import *

class playlist:
    """ a playlist is comprised of a number of objects. these objects can have
    arbitary elements within them. """
    
    def __init__(self):
        # this must be set here, first, or things will break - warned you! :)
        self.__dict__['callbacks'] = []
        self.__dict__['locked']=[]
        self.dirty=0
        self.list = []
        self.cur = 0
        self.last_jump = None

    def append(self, obj):
        self.list.append(obj)

    def add_hook(self, hook):
        """ add a hook to be called when any inst var is changed """
        self.callbacks.append(hook)

    def __setattr__(self, name, value):
        """ when an instance var is changed, call our hooks """
        self.__dict__[name] = value
        if not self.locked:
            [x() for x in self.callbacks]
        self.__dict__['dirty']=1

    def __getitem__(self, index):
        return self.list[index]

    def lock(self, l):
        """ Enable/disable the calling of playlist hooks while updating """
        
        if l:
            self.__dict__['locked']=1
        else:
            # triggers the hooks again
            self.locked=0

    def push(self):
        """ return the first item in the list, chopping it off """

        if len(self.list) == 0:
            return None

        return self.list.pop(0)

    def clear(self):
        self.list = []
        self.cur = 0

    def __len__(self):
        return len(self.list)
        
    def set_pos(self, pos):
        if pos < 0 or pos > len(self): raise ValueError("index out of bounds")
        self.cur = pos

    def get_pos(self):
        """ return the current playlist position, if any """

        if len(self) == 0:
            return

        return self.cur

    def next(self):
        """ advance one position in the playlist. loop if we've hit the end.
        returns 1 on normal skip, 2 on playlist loop, and 0 if the playlist
        is empty """

        if len(self) == 0:
            return 0

        if self.cur >= len(self)-1:
            self.cur = 0
            return 2
        else:
            self.cur += 1
            return 1

    def prev(self):
        """ skip back one position in the playlist. loop if start again
        returns 1 on normal skip, 2 on playlist loop, and 0 if the playlist
        is empty """
        
        if len(self) == 0:
            return 0

        if self.cur <= 0:
            self.cur = len(self) - 1
            return 2
        else:
            self.cur -= 1
            return 1

    def jump(self, amount):
        """ jump forward or back through the playlist a given number of times.
        the sign of amount determines the direction to proceed. wrapping will
        occur where appropriate """

        if amount > 0:
            for i in range(amount):
                self.next()
        else:
            for i in range(-amount):
                self.prev()

    def jump_to_path(self, path):
        """ attempt to jump to a file. return 0 on failure, 1 on success. """

        for x in range(len(self.list)):
            if self.list[x]['path'] == path:
                self.set_pos(x)
                return 1

        return 0

    def current(self, keepdirty=1):
        """ return the 'current' object """

        # no-op if we're empty
        if len(self) == 0:
            return

        return self.list[self.cur]

    def sort(self):
        """ sort the current playlist """

        def by(key, a, b):
            """ -1 if a < b, 0 ==, 1 if a > b """
            try:
                x = a[key]
                y = b[key]
            except KeyError:
                return 0

            if x == y: return 0
            if x > y: return 1
            if x < y: return -1            

        def all(a, b, by=by):
            foo = by('dir', a, b)
            if foo==0:
                foo = by('playlist', a, b)
                if foo==0:                    
                    foo = by('track', a, b)
                    if foo==0:
                        foo = by('artist', a, b)
                        if foo==0:
                            return by('title', a, b)
                        else: return foo
                    else: return foo
                else: return foo                        
            else: return foo
            

        self.list.sort(all)

    def reverse(self):
        """ reverse the order of the current playlist """
        self.list.reverse()

        # inline rev won't trigger playlist hooks.
        self.list = self.list
        

    def next_artist(self):
        """ jump to the first song by the next artist in the playlist. not as
        effective on an unsorted list, but leave it up to the user. this is
        where the return code from next() et al comes in handy. without it,
        we'd loop indefinitely if there was only one artist."""

        if len(self.list) == 0: return

        current_artist = self.list[self.cur]['artist']

        self.lock(1)
        while self.list[self.cur]['artist'] == current_artist and \
              self.next() == 1: pass
        self.lock(0)


    def prev_artist(self):
        """ jump to the last song by the previous artist in the
        playlist. not as effective on an unsorted list, but leave it
        up to the user. this is where the return code from next() et
        al comes in handy. without it, we'd loop indefinitely if there
        was only one artist."""

        if len(self.list) == 0: return

        current_artist = self.list[self.cur]['artist']

        self.lock(1)
        while self.list[self.cur]['artist'] == current_artist and \
              self.prev() == 1: pass
        self.lock(0)

    def next_dir(self):
        """ jump to the first song in the next dir in the playlist. """
        
        if len(self.list) == 0: return

        cur_dir = self.list[self.cur]['dir']

        self.lock(1)
        while self.list[self.cur]['dir'] == cur_dir and \
              self.next() == 1: pass
        self.lock(0)        

    def prev_dir(self):
        """ jump to the last song in the previous dir in the playlist. """
        
        if len(self.list) == 0: return

        cur_dir = self.list[self.cur]['dir']

        self.lock(1)
        while self.list[self.cur]['dir'] == cur_dir and \
              self.prev() == 1: pass
        self.lock(0)

    def jump_to_field(self, field=None, string=None):

        if len(self.list) == 0: return

        c = self.cur + 1

        # the user would like the previous
        # search repeated
        if field == string == None:
            if self.last_jump == None:
                return 0
            (field, string) = self.last_jump
        else:
            self.last_jump = (field, string)

        # now see if we match the artist - if we've looped around,
        # then the artist couldn't be found, and we fail.

        while c != self.cur:

            if self.list[c][field].lower().find(string.lower())!=-1:
                self.set_pos(c)                
                return 1

            c += 1
            if c>=len(self.list): c = 0
            
        return 0

# global instances, functions, etc
list = playlist()
queue = playlist()

# testing

if __name__ == '__main__':

    print "Very-ultra-incredibly-simplistic tests started."

    p = playlist()

    p.append("file1")
    p.append("file2")
    p.append("file3")

    p.next()
    assert p.current() == "file2"

    p.jump(-2)
    assert p.current() == "file3"

    p.prev()
    assert p.current() == "file2"

    print "Tests passed OK!"


    
        
    
