#!/usr/bin/env python

import curses
import sys
import os # FIXME, can be removed later
import atexit

import ui
import multiplex
import player
import playlist

from debug import debug

PLUGIN_TYPE="ui"

class new(ui.plugin):
    """ provides an abstract class for a interface plugin, for muse"""

    def _onexit(self):
        """ called by atexit, clean up the screen and prevent further
        modifications """
        self.stdscr.clear()
        self.stdscr.refresh()
        self.stdscr.keypad(0)
        self.stdscr.leaveok(0)
        curses.nocbreak()
        curses.echo()
        curses.endwin()

        # evil hack to prevent modules from writing to the screen after we close
        ui.refresh = lambda: 1

    def __init__(self):

        ui.plugin.__init__(self)

        self.stdscr = curses.initscr()

        curses.noecho()
        curses.cbreak()
        self.stdscr.keypad(1)
        self.stdscr.leaveok(1)
        self.stdscr.clear()

        # setup exit handler
        atexit.register(self._onexit)

        # now lay out the screen as it should appear

        # clear top line
        self.stdscr.move(0, 0)
        self.stdscr.attron(curses.A_REVERSE)
        self.stdscr.addstr(" " * curses.COLS)
        # clear bottom line
        self.stdscr.move(curses.LINES - 1, 0)
        self.stdscr.addstr(" " * (curses.COLS - 1)) 
        self.stdscr.attroff(curses.A_REVERSE)

        self.stdscr.refresh()

        fd = sys.stdin.fileno()
        mask = multiplex.POLLIN
        multiplex.add_watch(fd, mask, self.hook)

        playlist.list.add_hook(self.draw_playlist)

        mid = curses.LINES/2
        self.playlist_mid = mid
        self.playlist_top = mid - (mid-2)
        self.playlist_bot = mid + (mid-2)
            
        if curses.LINES % 2 != 0:
            # odd number of lines, add one to bottom
            self.playlist_bot += 1

        ui.add_callback('song', self.set_song)
        ui.add_callback('status', self.set_status)
        ui.add_callback('notice', self.notice)
        ui.add_callback('time', self.set_time)

        # evil, but curses doesn't seem to do it for me?
        # os.system("(sleep 2; tput civis)&")

    def hook(self, fd):
        """ called when there's action on the FD we're watching - ie keypress,
        etc """

        # okay, there's action on the keyboard. read the key and parse the
        # entry.

        # if you're wondering about the weird keybindings, i use a dvorak
        # keyboard :-)

        c = self.stdscr.getch()

        if c >= 0 and c < 256:
            char = chr(c)
        else:
            char = None

        # backspace: next song in queue
        if c == 263: 
            player.next_with_queue()

        # tab: add current to queue
        elif c == 9:
            player.add_to_queue()
            ui.notice("added to queue - %d songs waiting" % \
                      (len(playlist.queue)-1))
            playlist.list.next()
            
        elif char == "f":
            player.fwd()
        elif char == "b":
            player.back()

        elif char == " ":
            player.next()

        elif char == 'z':
            player.prev()

        elif char == "`": 
            player.play_current()
            ui.notice("playing current")

        elif char == '\n':
            player.next_artist()

        elif char == '\\':
            player.prev_artist()

        elif char == '-':
            player.next_dir()

        elif char == '=':
            player.prev_dir()

        elif char == 'o' or c == 260:
            playlist.list.prev_artist()
            ui.notice("jumping to prev artist")

        elif char == 'q' or c == 261:
            playlist.list.next_artist()
            ui.notice("jumping to next artist")

        elif char == 'e':
            playlist.list.prev_dir()
            ui.notice("prev dir: " + \
                      os.path.split(playlist.list.current()['dir'])[1])
                      
        elif char == 'j':
            playlist.list.next_dir()
            ui.notice("next dir: " + \
                      os.path.split(playlist.list.current()['dir'])[1])

        elif c == 265:
            player.start(persist=0)

        elif char == "~":
            player.message("switch_player")

        elif c == 12: # C-l
            self.stdscr.clearok(1)
            ui.refresh()

        elif char == 'a' or c == 259:
            playlist.list.prev()

        elif char == ';' or c == 258:
            playlist.list.next()                

        elif char == 'Q':
            player.quit()

        elif char == 'p':
            player.pause()


    def blank(self):
        """ blank - toggle the 'screen' on and off. doesn't need to be
        implemented for all interface modules"""
        pass


    def notice(self, msg):
        """ shows a low-priority message - the sort of thing to appear briefly
        in a status bar, for instance."""

        msg = msg[:(curses.COLS-2)] # strip to make sure it'll fit

        self.stdscr.move(curses.LINES-1, 0)

        self.stdscr.attron(curses.A_REVERSE)
        self.stdscr.addstr(" " * (curses.COLS-1))

        self.stdscr.addstr(curses.LINES-1, 1, msg, curses.A_REVERSE)

        self.stdscr.attroff(curses.A_REVERSE)        

        # now we add a timer to clear the notice in ~ 1.5 seconds
        multiplex.set_timer("curses_notice_timer", 2, \
                            lambda x=self: x.notice(""))

    def set_time(self):
        """given a time in seconds, update the display"""

        secs = player.secs()

        self.stdscr.move(0, curses.COLS - 7)

        self.stdscr.attron(curses.A_REVERSE)
        self.stdscr.addstr(" " * 7)

        if secs != None:
            str = "%.2d:%.2d" % (int(secs/60),secs%60)
        else:
            str = "     "
        
        self.stdscr.addstr(0, curses.COLS-len(str)-1, str)

        self.stdscr.attroff(curses.A_REVERSE)

    def set_song(self):

        if player.playlist_cur != None:
            song = player.playlist_cur['artist']

            if player.playlist_cur['title']:
                song += ' - ' + player.playlist_cur['title']
        else:
            song = ""

        self.stdscr.move(0,0)
        self.stdscr.attron(curses.A_REVERSE)
        self.stdscr.addstr(" " * (curses.COLS - 15)) # save 15 for time, etc

        # crop string
        song = song[: curses.COLS - 14]

        # add it

        self.stdscr.addstr(0,1, song)

        self.stdscr.attroff(curses.A_REVERSE)

    def set_status(self):

        status = player.status()
        
        self.stdscr.move(0, curses.COLS - 15)
        self.stdscr.attron(curses.A_REVERSE)
        self.stdscr.addstr(" " * 9) # save 15 for time, etc

        # crop string
        status = status[:8]

        # add it

        self.stdscr.addstr(0,curses.COLS - 15, status)

        self.stdscr.attroff(curses.A_REVERSE)        

    def refresh(self):
        self.stdscr.refresh()

    def playlist_line(self, y, song, hi=0):
        self.stdscr.move(y, 1)

        song = song[:curses.COLS-3]
        
        if hi:
            self.stdscr.attron(curses.A_REVERSE)
            # pad to end of line
            song = song.ljust(curses.COLS-3)

        self.stdscr.addstr(song)

        if hi:
            self.stdscr.attroff(curses.A_REVERSE)

#                player.clear_queue()
#                player.play_next()
                 #player.next()
        self.stdscr.clrtoeol()

    def draw_playlist(self, queue=0):
        """ draw the normal playlist. if passed with queue=1, draw the queue
        instead """

        #debug("redraw playlist with pl_len=" + str(len(playlist.list)))

        if len(playlist.list)==0:
            self._clear_playlist()            
            return

        old_dir = None

        cur = playlist.list.cur
        top = cur - (self.playlist_mid - self.playlist_top)
        # for x in range(0, cur - self.playlist_top, cur + self.playlist_bot):
        for x in range(self.playlist_top, self.playlist_bot):
            if top >= 0 and top < len(playlist.list):

                song = playlist.list[top]['artist'] + " - "

                if playlist.list[top].has_key('track'):
                    foo = str(playlist.list[top]['track'])
                    if len(foo)==1: foo = "0" + foo
                    song = song + foo + '. '

                song += playlist.list[top]['title']

                if x == self.playlist_mid:
                    self.playlist_line(x, song, 1)
                else:
                    self.playlist_line(x, song)
            else:
                self.playlist_line(x, "")

            top += 1
        ui.refresh()

    def draw_artist(self):
        """ draw a list of the artist available to jump to """

        if len(playlist.list)==0:
            self._clear_playlist()
            return

        cur = playlist.list.cur
        top = cur - (self.playlist_mid - self.playlist_top)

        artist = playlist.current()['artist']
        self.playlist_line(x, artist, 1)

        for x in range(self.playlist_top, self.playlist_bot):
            pass
            

        

    def _clear_playlist(self):
        for x in range(self.playlist_top, self.playlist_bot):
            self.stdscr.move(x, 0)
            self.stdscr.clrtoeol()
            
        ui.refresh()

        
        


    
    
        
        

    
        
    
